# Mock HTCondor library for CI testing
# Developed with assistance from ChatGPT by OpenAI
# Provides simulated functionality for HTCondor, including
# job submission, querying, and management
from collections import defaultdict
import htcondor
import logging

log = logging.getLogger('falconry')


class MockSchedd:
    def __init__(self):
        self.job_queue = defaultdict(dict)
        self.job_history = defaultdict(dict)
        self.job_id_counter = 1
        self.log_files = {}  # Simulate log files per job

    def submit(self, job_description, itemdata=[]):
        """Simulates the submission of a job."""
        if itemdata == []:
            itemdata = [None]
        for i, proc in enumerate(itemdata):
            job_id = f"{self.job_id_counter}.{i}"

            # Replace $(JobId) in the log file path
            log_file_path = job_description.get("Log", "mock_condor_$(JobId).log")
            log_file_path = log_file_path.replace("$(JobId)", str(job_id))

            # Create an initial log file
            self._write_log_file(log_file_path, "Job is idle.")

            self.job_queue[job_id] = {
                "JobDescription": job_description,
                "JobStatus": MockHTCondor.job_status_map()["Idle"],
            }
        self.job_id_counter += 1
        return MockSubmitResult(self.job_id_counter - 1)

    def get_constraint(self, constraint):
        if constraint is None:
            return lambda job_id, job_info: True
        else:
            if "==" in constraint and 'ClusterId' in constraint:
                constraint = constraint.split("&&")[0].split("==")

                def cnstr(job_id, job_info):
                    print(constraint[1].strip().replace(")", ""))
                    return job_id.split(".")[0] == constraint[1].strip().replace(
                        ")", ""
                    )

                return cnstr

    def query(self, constraint=None, projection=None):
        """Simulates querying the job queue."""
        # interpret constraint
        # for now only ==
        # TODO: add more
        l_constraint = self.get_constraint(constraint)

        result = []
        for job_id, job_info in self.job_queue.items():
            if l_constraint(job_id, job_info):
                if projection:
                    result.append({key: job_info.get(key, None) for key in projection})
                else:
                    result.append(job_info)
        return result

    def edit(self, job_id, attribute, value):
        """Simulates editing an attribute of a job."""
        if job_id in self.job_queue:
            self.job_queue[job_id][attribute] = value
        else:
            raise ValueError(f"Job ID {job_id} not found.")

    def remove(self, job_id):
        """Simulates removing a job from the queue."""
        if job_id in self.job_queue:
            job_info = self.job_queue[job_id]
            job_description = job_info["JobDescription"]
            log_file_path = job_description["Log"]
            log_file_path = log_file_path.replace("$(JobId)", str(job_id))
            self._write_log_file(log_file_path, "Job was aborted by the user.")
            self.job_history[job_id] = job_info
            del self.job_queue[job_id]
        else:
            raise ValueError(f"Job ID {job_id} not found.")

    def run_jobs(self):
        """Simulates running all idle jobs."""
        for job_id, job_info in self.job_queue.items():
            if job_info["JobStatus"] == MockHTCondor.job_status_map()["Idle"]:
                job_info["JobStatus"] = MockHTCondor.job_status_map()["Running"]
                log_file_path = job_info["JobDescription"]["Log"]
                log_file_path = log_file_path.replace("$(JobId)", str(job_id))
                self._write_log_file(log_file_path, "Job is running.")

    def complete_jobs(self):
        """Simulates completing all running jobs."""
        to_delete = []
        for job_id, job_info in self.job_queue.items():
            if job_info["JobStatus"] == MockHTCondor.job_status_map()["Running"]:
                job_info["JobStatus"] = MockHTCondor.job_status_map()["Completed"]
                log_file_path = job_info["JobDescription"]["Log"]
                log_file_path = log_file_path.replace("$(JobId)", str(job_id))
                self._write_log_file(
                    log_file_path, "Job terminated\nNormal termination (return value 0)"
                )
                self.job_history[job_id] = job_info
                to_delete.append(job_id)
        for job_id in to_delete:
            del self.job_queue[job_id]

    def fail_job(self, job_id, fail_code):
        """Simulates failing a specific job with a custom failure code."""
        if job_id in self.job_queue:
            job_info = self.job_queue[job_id]
            if job_info["JobStatus"] == MockHTCondor.job_status_map()["Running"]:
                job_info["JobStatus"] = MockHTCondor.job_status_map()["Completed"]
                log_file_path = job_info["JobDescription"]["Log"]
                log_file_path = log_file_path.replace("$(JobId)", str(job_id))
                self._write_log_file(
                    log_file_path,
                    f"Job terminated\nNormal termination (return value {fail_code})",
                )
                self.job_history[job_id] = job_info
                del self.job_queue[job_id]
        else:
            raise ValueError(f"Job ID {job_id} not found.")

    def act(self, action, constraint):
        """Simulates performing an action on jobs based on a constraint."""
        l_constraint = self.get_constraint(constraint)
        for job_id, job_info in list(self.job_queue.items()):
            if l_constraint(job_id, job_info):
                if action == htcondor.JobAction.Release:
                    if job_info["JobStatus"] == MockHTCondor.job_status_map()["Held"]:
                        job_info["JobStatus"] = MockHTCondor.job_status_map()["Idle"]
                        log_file_path = job_info["JobDescription"]["Log"]
                        log_file_path = log_file_path.replace("$(JobId)", str(job_id))
                        self._write_log_file(log_file_path, "Job is idle.")
                elif action == htcondor.JobAction.Remove:
                    self.remove(job_id)
                # Add more actions as needed

    def history(self, constraint=None, projection=None):
        """Simulates retrieving the history of completed jobs."""
        result = []
        l_constraint = self.get_constraint(constraint)
        for job_id, job_info in self.job_history.items():
            if job_info["JobStatus"] == MockHTCondor.job_status_map()["Completed"]:
                if l_constraint(job_id, job_info):
                    if projection:
                        result.append(
                            {key: job_info.get(key, None) for key in projection}
                        )
                    else:
                        result.append(job_info)
        return result

    def _write_log_file(self, log_file_path, content):
        """Writes content to the specified log file."""
        if log_file_path:
            with open(log_file_path, "w") as log_file:
                log_file.write(content)


class MockSubmitResult:
    def __init__(self, cluster_id):
        self.cluster_id = cluster_id

    def cluster(self):
        """Returns the cluster ID of the submitted job."""
        return self.cluster_id

    def __str__(self):
        return f"{self.cluster_id}"


class MockHTCondor:
    @staticmethod
    def Schedd():
        """Returns a mock Schedd object."""
        return MockSchedd()

    @staticmethod
    def job_status_map():
        """Provides a mapping of job statuses."""
        return {
            "Idle": 1,
            "Running": 2,
            "Completed": 4,
            "Removed": 3,
            "Held": 5,
        }



import sys
import select
import logging
from enum import Enum
from typing import Dict, Optional, Tuple

log = logging.getLogger('falconry')


# helper class to define the status
class InputState(Enum):
    UNKNOWN = -1
    SUCCESS = 0
    TIMEOUT = 1


def input_checker(
    validOptions: Dict[str, str],
    timeout: int = 60,
    message: str = "Following options available:",
    silent: bool = False,
) -> Tuple[InputState, Optional[str]]:
    """Helper function to get input from user

    Arguments:
        validOptions (Dict[str, str]): dictionary of valid options
        timeout (int, optional): timeout in seconds. Defaults to 60.
        message (str, optional): message to print before options.
        Defaults to "Following options available:".
        silent (bool, optional): silent mode. Defaults to False.

    Returns:
        Tuple[InputState, Optional[str]]: returns the state of the input
    """
    if message != "" and not silent:
        log.info(message)
    for opt, desc in validOptions.items():
        if desc != "" and not silent:
            log.info(f"{opt} - {desc}")

    i, o, e = select.select([sys.stdin], [], [], timeout)
    if i:
        inp = sys.stdin.readline().strip()
        if inp in validOptions.keys():
            return InputState.SUCCESS, inp

        if not silent:
            log.info(f"Unknown state {inp}!")
        return InputState.UNKNOWN, inp

    else:
        if not silent:
            log.info("Timed out ...")
        return InputState.TIMEOUT, None



import htcondor  # for submitting jobs, querying HTCondor daemons, etc.
import logging
import functools
from typing import Callable, Any
import time

log = logging.getLogger('falconry')


# Here the typing did not work properly ...
def schedd_check(func: Callable[["ScheddWrapper"], Any]) -> Any:
    @functools.wraps(func)
    def wrapper(
        self: "ScheddWrapper", *args: Any, **kwargs: Any
    ) -> Callable[["ScheddWrapper"], Any]:
        try:
            return func(self, *args, **kwargs)
        except htcondor.HTCondorIOError:
            log.warning(
                "Possible problem with scheduler, waiting a bit and reloading schedd ..."
            )
            time.sleep(60)
            self.schedd = htcondor.Schedd()
            return func(self, *args, **kwargs)

    return wrapper


def kerberos_auth() -> None:
    """Add kerberos creds to the schedd"""
    try:
        credd = htcondor.Credd()
        credd.add_user_cred(htcondor.CredTypes.Kerberos, None)
    except:
        log.warning(
            "Kerberos creds not available. This can cause problems on some clusters (like lxplus)."
        )


class ScheddWrapper:
    """Wrapper to allow reload of schedd"""

    def __init__(self) -> None:
        self.schedd = htcondor.Schedd()

    """Reimplementing all the used functions"""

    @schedd_check
    def act(self, *args: Any, **kwargs: Any) -> int:
        return self.schedd.act(*args, **kwargs)

    @schedd_check
    def query(self, *args: Any, **kwargs: Any) -> list[dict[str, Any]]:
        return self.schedd.query(*args, **kwargs)

    @schedd_check
    def history(self, *args: Any, **kwargs: Any) -> list[dict[str, Any]]:
        return self.schedd.history(*args, **kwargs)

    @schedd_check
    def submit(self, *args: Any, **kwargs: Any) -> htcondor.SubmitResult:
        return self.schedd.submit(*args, **kwargs)



# Test job.py using the htcondor_mock library
from MockHTCondor import MockHTCondor
from falconry import job, manager, Counter, FalconryStatus
import pytest


def test_job():
    schedd = MockHTCondor.Schedd()
    j = job("test", schedd)  # type: ignore
    j.set_simple("my_script.sh", "log")
    assert j.name == "test"
    assert j.schedd == schedd
    assert j.jobIDs == []
    assert j.release() is False
    assert j.remove() is False
    assert j.submitted is False
    assert j.done is False
    assert j.skipped is False

    with pytest.raises(SystemError):
        j.get_info()

    j.submit()
    assert j.submitted is True
    assert len(j.jobIDs) == 1
    assert j.jobIDs[0] == "1.0"
    assert j.jobID == "1.0"

    assert j.remove() is True
    assert j.get_info()["JobStatus"] == -999
    assert j.get_status() == FalconryStatus.ABORTED_BY_USER

    j.submit()
    assert j.submitted
    assert len(j.jobIDs) == 2
    assert j.jobIDs[0] == "1.0"
    assert j.jobIDs[1] == "2.0"
    assert j.jobID == "2.0"
    assert j.get_status() == FalconryStatus.IDLE
    assert j.get_info()["JobStatus"] == j.get_status().value

    schedd.run_jobs()

    assert j.get_status() == FalconryStatus.RUNNING
    assert j.get_info()["JobStatus"] == j.get_status().value

    schedd.complete_jobs()

    assert j.get_status() == FalconryStatus.COMPLETE
    assert j.get_info()["JobStatus"] == j.get_status().value


def test_manager():
    schedd = MockHTCondor.Schedd()
    mgr = manager("log", schedd=schedd)  # type: ignore

    j = job("test", schedd)  # type: ignore
    j.set_simple("my_script.sh", "log")
    mgr.add_job(j)

    c = Counter()

    assert mgr._single_check(c) is True
    mgr._submit_jobs()
    assert j.get_status() == FalconryStatus.IDLE
    schedd.run_jobs()
    assert mgr._single_check(c) is True
    assert j.get_status() == FalconryStatus.RUNNING
    schedd.complete_jobs()
    assert mgr._single_check(c) is False
    assert j.get_status() == FalconryStatus.COMPLETE
    mgr.save(quiet=True)
    mgr.load(retryFailed=False)


if __name__ == "__main__":
    test_job()
    test_manager()



#!/usr/bin/env python3

import logging
from falconry import manager, job

logging.basicConfig(level=logging.INFO, format="%(levelname)s (%(name)s): %(message)s")
log = logging.getLogger(__name__)


# using argparse to get command line arguments
def config():
    import argparse

    parser = argparse.ArgumentParser(description="Falconry. Read README!")
    parser.add_argument(
        "--debug", action="store_true", help="Debug mode with more verbose printout"
    )
    parser.add_argument(
        "--dir",
        type=str,
        help="Path to output DIR. Output by default.",
        default="Output",
    )
    return parser.parse_args()


def main():

    # Getting command line arguments and creating the manager
    # the manager will periodically check the jobs and handle dependencies
    cfg = config()

    if cfg.debug:
        logging.getLogger("falconry").setLevel(logging.DEBUG)

    mgr = manager(cfg.dir)  # the argument specifies where the job is saved

    # It is alway useful to save the printounts to a log file
    file_handler = logging.FileHandler(cfg.dir + "/falconry.log", mode="a")
    dt_fmt = '%Y-%m-%d %H:%M:%S'
    formatter = logging.Formatter(
        '[{asctime}] [{levelname:<8}] {name}: {message}', dt_fmt, style='{'
    )
    file_handler.setFormatter(formatter)
    log.addHandler(file_handler)

    # Check if manager was already run in given dir and ask user
    # if they want to load previous instance
    # First return value if user typed reasonable input (quite if not)
    # Second is the actual value, possible are 'l' to load
    # and 'n' for new right now
    load = False
    status, var = mgr.check_savefile_status()
    if status:
        if var == "l":
            load = True
    else:
        return

    # Ask for message to be saved in the save file
    # Always good to have some documentation ...
    mgr.ask_for_message()

    # Short function to simplify the job setup
    def simple_job(name: str, exe: str) -> job:
        # define job and pass the HTCondor schedd to it
        j = job(name, mgr.schedd)

        # set the executable and the path to the log files
        j.set_simple(exe, cfg.dir + "/log/")

        # set the expected run time in seconds
        j.set_time(120)
        return j

    if load:
        # Continue where we left ...
        mgr.load()
    else:
        # As a test, prepare one job which succeeds and one that fails
        j = simple_job("success", "util/echoS.sh")
        # use job::submit to submit the job. With manager, this is not necessary,
        # when job does not have and dependency it is submitted automatically!
        j.submit()
        mgr.add_job(j)
        depS = [j]

        j = simple_job("error", "util/echoE.sh")
        # Here we let manager to handle the submition
        # They should be submitted automatically
        # to the same cluster
        mgr.add_job(j)
        depE = [j]
        j = simple_job("error2", "util/echoE.sh")
        mgr.add_job(j)
        depE.append(j)

        # add job depending on the submitted job to demonstrate what happens if depending job fails or succeds
        j = simple_job("success_depS", "util/echoS.sh")
        j.add_job_dependency(*depS)
        mgr.add_job(j)

        j = simple_job("success_depE", "util/echoS.sh")
        j.add_job_dependency(*depE)
        mgr.add_job(j)

    # Start the manager
    # If there is an error, especially interupt with keyboard,
    # it saves the current state of jobs
    mgr.start(60, gui=False)  # argument is interval between checking of the jobs
    # Save the final status and print failed jobs
    mgr.save()
    mgr.print_failed()


if __name__ == "__main__":
    main()



import logging
import json
import ijson
import os
import shutil
import sys
import traceback
import datetime
import select
from time import sleep
import htcondor

from typing import Dict, Any, Tuple, Optional

from .job import job
from .status import FalconryStatus
from . import cli
from .schedd_wrapper import ScheddWrapper

log = logging.getLogger('falconry')


class Counter:
    # just holds few variables used in status print
    def __init__(self) -> None:
        self.waiting = 0
        self.notSub = 0
        self.idle = 0
        self.run = 0
        self.failed = 0
        self.done = 0
        self.skipped = 0
        self.removed = 0
        self.held = 0

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Counter):
            return NotImplemented
        return (
            self.waiting == other.waiting and
            self.notSub == other.notSub and
            self.idle == other.idle and
            self.run == other.run and
            self.failed == other.failed and
            self.done == other.done and
            self.skipped == other.skipped and
            self.removed == other.removed and
            self.held == other.held
        )

class manager:
    """Manager holds all jobs and periodically checks their status.

    It also take care of dependent jobs,
    submitting jobs when all dependencies are satisfied.
    These are handled as decorations of the job.

    Arguments:
        mgrDir (str): directory where the manager stores the jobs
        mgrMsg (str): message to be saved in the save file
        maxJobIdle (int): maximum number of idle jobs
    """

    reservedNames = ["Message", "Command"]

    def __init__(
        self,
        mgrDir: str,
        mgrMsg: str = "",
        maxJobIdle: int = -1,
        schedd: Optional[ScheddWrapper] = None,
    ):
        log.info("MONITOR: INIT")

        # Initialize the manager, maily getting the htcondor schedd
        if schedd is not None:
            self.schedd = schedd
        else:
            self.schedd = ScheddWrapper()

        # job collection
        self.jobs: Dict[str, job] = {}
        self.sub_queue: list[job] = []

        # now create a directory where the info about jobs will be save
        if not os.path.exists(mgrDir):
            os.makedirs(mgrDir)
        self.dir = mgrDir
        self.saveFileName = self.dir + "/data.json"
        self.mgrMsg = mgrMsg
        self.command = " ".join(sys.argv)

        self.maxJobIdle = maxJobIdle
        self.curJobIdle = 0

    # check if save file already exists
    def check_savefile_status(self) -> Tuple[bool, Optional[str]]:
        """Checks if the save file already exists. If it does, asks the user
        whether to load existing jobs or start new ones.

        Returns:
            Tuple[bool, Optional[str]]: (True, 'l') if load, (True, 'n') if new,
            (False, None) if error
        """
        if os.path.exists(self.saveFileName):
            log.warning(f"Manager directory {self.dir} already exists!")
            state, var = cli.input_checker(
                {"l": "Load existing jobs", "n": "Start new jobs"}
            )

            # Simplify the output for user interface
            # both unknown/timeout have the same result
            if state == cli.InputState.SUCCESS:
                return True, var
            return False, var
        return True, "n"  # automatically assume new

    def ask_for_message(self) -> None:
        """Asks user for a message to be saved in the save file for bookkeeping."""

        log.info("Enter a message to be saved in the save file " "for bookkeeping.")
        i, o, e = select.select([sys.stdin], [], [], 60)
        if i:
            self.mgrMsg = sys.stdin.readline().strip()

    def add_job(self, j: job, update: bool = False) -> None:
        """Adds a job to the manager. If the job already exists and `update` is
        `True`, it will be updated.

        Arguments:
            j (job): job to be added
            update (bool, optional): whether to update the job if it already
            exists. Defaults to False.
        """
        # some reserved names, to simplify saving later
        if j.name in manager.reservedNames:
            log.error("Name %s  is reserved! Exiting ...", j.name)
            raise SystemExit

        # first check if the jobs already exists
        if j.name in self.jobs.keys():
            if not update:
                log.error("Job %s already exists! Exiting ...", j.name)
                raise SystemExit
            else:
                log.info(f"Updating job {j.name}.")

        self.jobs[j.name] = j

    def save(self, quiet: bool = False) -> None:
        """Saves the current status of the jobs to a json file.

        If `quiet` is `True`, it will not print any messages and
        will not make a time-stamped copy of the save file.

        Arguments:
            quiet (bool, optional): whether to print messages. Defaults to False.
        """
        if not quiet:
            log.info("Saving current status of jobs")
        output: Dict[str, Any] = {
            "Message": self.mgrMsg,
            "Command": self.command,
        }
        for name, j in self.jobs.items():
            output[name] = j.save()

        # save with a timestamp as a suffix, create sym link
        current_time = datetime.datetime.now().strftime("%Y%m%d_%H%M_%S")
        fileLatest = f"{self.saveFileName}.latest"
        fileSuf = f"{self.saveFileName}.{current_time}"  # only if not quiet

        with open(fileLatest, "w") as f:
            json.dump(output, f, indent=2)
            if not quiet:
                log.info("Success! Making copy with time-stamp.")
                if not os.path.exists(fileSuf):
                    shutil.copyfile(fileLatest, fileSuf)
                else:
                    raise FileExistsError(
                        f"Destination file {fileSuf} already exists. "
                        "This should not be possible."
                    )

        # not necessary to remove, but maybe better to be sure its not broken
        if os.path.exists(self.saveFileName):
            os.remove(self.saveFileName)
        os.symlink(fileLatest.split("/")[-1], self.saveFileName)

    def load(self, retryFailed: bool = False) -> None:
        """Loads the saved status of the jobs from a json file
        provided by the user.

        Arguments:
            retryFailed (bool, optional): whether to retry the failed jobs.
            Defaults to False.
        """
        log.info("Loading past status of jobs")
        with open(self.dir + "/data.json", "rb") as f:
            depNames = {}
            for name, jobDict in ijson.kvitems(f, ""):
                if name in manager.reservedNames:
                    continue
                log.debug("Loading job %s", name)

                # create a job
                j = job(name, self.schedd)
                j.load(jobDict)

                # add it to the manager
                self.add_job(j, update=True)

                # decorate the list of names of the dependencies
                depNames[j.name] = jobDict["depNames"]

        # Now that jobs are defined, dependencies can be recreated
        # also resubmit jobs which failed
        for j in self.jobs.values():
            dependencies = [self.jobs[name] for name in depNames[j.name]]
            j.add_job_dependency(*dependencies)

        # Retry failed jobs
        # Since this changes the status and submits
        # jobs, add safequard in case of crash
        # to save up-to-date state
        if retryFailed:
            try:
                for j in self.jobs.values():
                    self._check_resubmit(j, True)
            except KeyboardInterrupt:
                log.error("Manager interrupted with keyboard!")
                log.error("Saving and exitting ...")
                self.save()
                self.print_failed()
                sys.exit(0)
            except Exception:
                log.error("Error ocurred when running manager!")
                traceback.print_exc(file=sys.stdout)
                self.save()
                self.print_failed()
                sys.exit(1)

    # print names of all failed jobs
    def print_running(self, printLogs: bool = False) -> None:
        log.info("Printing running jobs:")
        for name, j in self.jobs.items():
            if j.get_status() == FalconryStatus.RUNNING:
                log.info("%s (id %u)", name, j.jobID)
                if printLogs:
                    log.info(f"log: {j.logFile}")
                    log.info(f"out: {j.outFile}")
                    log.info(f"err: {j.errFile}")

    # print names of all failed jobs
    def print_failed(self, printLogs: bool = False) -> None:
        """Prints names of all failed jobs.

        Arguments:
            printLogs (bool, optional): whether to print paths to logs.
                Defaults to False.
        """
        log.info("Printing failed jobs:")
        for name, j in self.jobs.items():
            if j.get_status() == FalconryStatus.FAILED:
                log.info(f"{name} (id {j.jobID})")
                if printLogs:
                    log.info(f"log: {j.logFile}")
                    log.info(f"out: {j.outFile}")
                    log.info(f"err: {j.errFile}")
        # TODO: maybe separate failed and removed?
        log.info("Printing removed jobs:")
        for name, j in self.jobs.items():
            if j.get_status() == FalconryStatus.REMOVED:
                log.info(f"{name} (id {j.jobID})")
                if printLogs:
                    log.info(f"log: {j.logFile}")
                    log.info(f"out: {j.outFile}")
                    log.info(f"err: {j.errFile}")

    def _check_dependence(self) -> None:
        """Checks if all dependencies of a job are done. If so, it will submit
        the job. If any of the dependencies failed, it will add the job to the
        skipped list.
        """

        # TODO: consider if not submitted jobs in a special list
        for name, j in self.jobs.items():
            # only check jobs which are neither submitted nor skipped
            if j.submitted or j.skipped:
                continue

            # if ready submit, single not done dependency leads to isReady=False
            isReady = True
            for tarJob in j.dependencies:
                # if any job is not done, do not submit
                if tarJob.done:
                    continue

                isReady = False

                if tarJob.skipped or tarJob.failed:
                    log.error(
                        f"Job {name} depends on job {tarJob.name} which either failed or was skipped! Skipping ..."
                    )
                    j.skipped = True

                status = tarJob.get_status()
                if status == FalconryStatus.REMOVED:
                    log.error(
                        f"Job {name} depends on job {tarJob.name} which is {FalconryStatus.REMOVED}! Skipping ..."
                    )
                    j.skipped = True

                break

            if isReady:
                # Check if we did not reach maximum number of submitted jobs
                if self.maxJobIdle != -1 and self.curJobIdle > self.maxJobIdle:
                    break  # break because it does not make sense to check any other jobs now
                j.submit(doNotSubmit=True)
                self.sub_queue.append(j)
                self.curJobIdle += 1  # Add the jobs as a idle for now

    def _check_resubmit(self, j: job, retryFailed: bool = False) -> None:
        """Checks if a job should be resubmitted due to some known problems.

        Arguments:
            j (job): job to check
            retryFailed (bool, optional): whether to also retry failed jobs.
                Defaults to False.
        """
        status = j.get_status()
        log.debug("Job %s has status %s", j.name, status.name)
        if status is FalconryStatus.ABORTED_BY_USER:
            log.warning(
                f"Error! Job {j.name} (id {j.jobID}) failed due to condor, rerunning"
            )
            j.submit(force=True, doNotSubmit=True)
            self.sub_queue.append(j)
        elif retryFailed and status is FalconryStatus.FAILED:
            log.warning(
                f"Error! Job {j.name} (id {j.jobID}) failed and will be retried, rerunning"
            )
            j.submit(force=True, doNotSubmit=True)
            self.sub_queue.append(j)
        elif retryFailed and status is FalconryStatus.REMOVED:
            log.warning(
                f"Error! Job {j.name} (id {j.jobID}) was removed and will be retried, rerunning"
            )
            j.submit(force=True, doNotSubmit=True)
            self.sub_queue.append(j)
        elif (
            retryFailed
            and j.submitted
            and (
                status
                in [FalconryStatus.NOT_SUBMITTED, FalconryStatus.LOG_FILE_MISSING]
            )
        ):
            log.warning(
                f"Error! Job {j.name} was not submitted succesfully (probably...), rerunning"
            )
            j.submit(force=True, doNotSubmit=True)
            self.sub_queue.append(j)
        elif retryFailed and j.skipped:
            log.warning(
                f"Error! Job {j.name} was skipped and will be retried, rerunning"
            )
            j.skipped = False

    def _count_jobs(self, counter: Counter) -> None:
        """Counts the number of jobs with different status.
        Resubmits jobs which failed due to condor problems.

        Arguments:
            c (counter): counter object to count the jobs
        """

        maxLength = 0
        for name, j in self.jobs.items():
            printStr = f"Checking {name}\t\t\t\t\t\t\r"
            if len(printStr) > maxLength:
                maxLength = len(printStr)
            print(printStr, end='')

            self._count_job(counter, j)

            print(" " * maxLength + "\r", flush=True, end='')

    def _count_job(self, c: Counter, j: job) -> None:
        """Updates the counter object with the status of a single job.
        Also resubmits jobs which failed due to condor problems.

        Arguments:
            c (counter): counter object to update
            j (job): job to check
        """

        # first check if job is not submitted, skipped or done
        if j.skipped:
            c.skipped += 1
            return
        if not j.submitted:
            c.waiting += 1
            return
        if j.done:
            c.done += 1
            return

        #  resubmit job which failed due to condor problems
        self._check_resubmit(j)

        # count job with different status
        status = j.get_status()
        if (
            status == FalconryStatus.NOT_SUBMITTED
            or status == FalconryStatus.LOG_FILE_MISSING
        ):
            c.notSub += 1
        elif status == FalconryStatus.IDLE:
            c.idle += 1
        elif status == FalconryStatus.RUNNING:
            c.run += 1
        elif status == FalconryStatus.FAILED:
            c.failed += 1
        elif status == FalconryStatus.COMPLETE:
            c.done += 1
        elif status == FalconryStatus.HELD:
            c.held += 1
        elif status == FalconryStatus.REMOVED:
            c.removed += 1

    def _submit_jobs(self) -> None:
        """Submits all jobs in the submission queue."""

        if len(self.sub_queue) == 0:
            return

        # First we need to group jobs with the same executable
        jobs_with_exe: Dict[str, list[job]] = {}
        for j in self.sub_queue:
            exe = j.config["executable"]
            if exe not in jobs_with_exe:
                jobs_with_exe[exe] = []
            # check that the log path is the same. This should be the case by default
            elif j.config["log"] != jobs_with_exe[exe][0].config["log"]:
                log.error("Jobs with same executable have different log paths")
                raise Exception
            jobs_with_exe[exe].append(j)

        ignore = ["executable", "log", "output", "error"]
        # Now we need to submit each group
        for exe, jobs in jobs_with_exe.items():
            #
            log.debug("Submitting %i jobs with executable %s", len(jobs), exe)

            job_pars = [x.config for x in jobs]
            all_pars = set(sum([list(x.keys()) for x in job_pars], []))

            def var_name(x: str) -> str:
                return f"VAR_{x}".replace("+", "p")

            pars_dict = {x: f'$({var_name(x)})' for x in all_pars if x not in ignore}
            job_pars_variable = []
            for par in job_pars:
                tmp = {}
                for k, v in par.items():
                    if k not in ignore:
                        tmp[var_name(k)] = v
                job_pars_variable.append(tmp)

            base_pars = {
                "executable": exe,
                "log": jobs[0].config["log"],
                "output": jobs[0].config["output"],
                "error": jobs[0].config["error"],
            }
            base_pars.update(pars_dict)
            base_submit = htcondor.Submit(base_pars)
            result = self.schedd.submit(base_submit, itemdata=iter(job_pars_variable))

            log.debug(f"Submitted cluster ID: {result.cluster()}")
            for it, j in enumerate(jobs):
                j.submit_done(result.cluster(), str(it))
        self.sub_queue = []

    def _start_cli(self, sleep_time: int = 60) -> None:
        """Starts the manager, iteratively checking status of jobs.

        Arguments:
            sleep_time (int, optional): time to sleep between checks.
                Defaults to 60.
        """
        # TODO: maybe add flag to save for each check? or every n-th check?

        log.info("MONITOR: START")

        c = Counter()
        event_counter = 0

        self._submit_jobs()

        while True:
            if not self._single_check(c):
                self._submit_jobs()
                break

            self._submit_jobs()

            # save with timestamp every 30 events
            # most important for first event when first
            # batch of jobs is defined
            if event_counter % 30 == 0:
                self.save()
            event_counter += 1

            if not self._cli_interface(sleep_time):
                break

        log.info("MONITOR: FINISHED")

    def _single_check(self, c: Counter) -> bool:
        """Single check in the manager loop.

        Returns:
            bool: True if manager should continue, False otherwise
        """
        log.info(
            f"|-Checking status of jobs [{datetime.datetime.now()}]----------------|",
        )

        cOld = c
        c = Counter()
        self._count_jobs(c)

        # if no job is waiting nor running, finish the manager
        if not (c.waiting + c.notSub + c.idle + c.run > 0):
            return False

        # only printout if something changed:
        if c != cOld:
            sleep(0.2)  # the printing sometimes breaks here, adding delay helps...
            log.info(
                "| nsub: {0:>4} | hold: {1:>5} | fail: {2:>6} | rem: {3:>6} | skip: {4:>5} |".format(
                    c.notSub, c.held, c.failed, c.removed, c.skipped
                )
            )
            log.info(
                "| wait: {0:>6} | idle: {1:>4} | RUN: {2:>5} | DONE: {3:>6} | TOT: {4:>6} |".format(
                    c.waiting, c.idle, c.run, c.done, len(self.jobs)
                )
            )

            # Update current idle of jobs managed by manager.
            # All new jobs submitted jobs in `check_dependence`
            # will increase this number, that why we create different
            # variable than `c.idle`
            self.curJobIdle = c.idle

            # checking dependencies and submitting ready jobs
            self._check_dependence()
            self.save(quiet=True)

            # instead of sleeping wait for input
            log.info(
                "|-Enter 'h' to show all commands, e.g. to resubmit or show failed jobs|"
            )

        return True

    def _cli_interface(self, sleep_time: int = 60) -> bool:
        """CLI interface for the manager.
\
        Possible commands:
            f: show failed jobs
            ff: show failed jobs and log paths
            s: save manager state
            r: show running jobs
            rr: show running jobs and log paths
            x: exit
            h: help


        Arguments:
            sleep_time (int, optional): time to sleep between checks.
                Defaults to 60.

        Returns:
            bool: True if manager should continue, False otherwise
        """
        print('>>>> ', end='', flush=True)
        state, var = cli.input_checker(
            {
                "h": "",
                "s": "",
                "f": "",
                "x": "",
                "ff": "",
                "retry all": "",
                "r": "",
                "rr": "",
            },
            silent=True,
            timeout=sleep_time,
        )
        if state == cli.InputState.TIMEOUT:
            print('\r    \r', end='', flush=True)
        elif state == cli.InputState.SUCCESS:
            if var == "f":
                self.print_failed()
            elif var == "s":
                self.save()
            elif var == "h":
                log.info(
                    "|-Enter 'f' to show failed jobs, 'ff' to also show log paths----------|"
                )
                log.info(
                    "|-Enter 'r' to show running jobs, 'rr' to also show log paths---------|"
                )
                log.info(
                    "|-Enter 'x' to exit, 's' to save or 'retry all' to retry all failed---|"
                )
                self._cli_interface(sleep_time)
            elif var == "ff":
                self.print_failed(True)
            elif var == "r":
                self.print_running()
            elif var == "rr":
                self.print_running(True)
            elif var == "x":
                log.info("MONITOR: EXITING")
                return False
            elif var == "retry all":
                for j in self.jobs.values():
                    self._check_resubmit(j, True)

        return True

    def _start_gui(self, sleepTime: int = 60) -> None:
        """Starts the manager with GUI, iteratively checking status of jobs.

        This is only experimental!

        Arguments:
            sleepTime (int, optional): time to sleep between checks.
                Defaults to 60.
        """

        log.warning("GUI version is only experimental!")
        import tkinter as tk

        window = tk.Tk()
        window.title("Falconry monitor")
        frm_counter = tk.Frame()

        def quick_label(name: str, x: int, y: int = 0) -> tk.Label:
            lbl = tk.Label(master=frm_counter, width=10, text=name)
            lbl.grid(row=y, column=x)
            return lbl

        quick_label("Not sub.:", 0)
        quick_label("Idle:", 1)
        quick_label("Running:", 2)
        quick_label("Failed:", 3)
        quick_label("Done:", 4)
        quick_label("Waiting:", 5)
        quick_label("Skipped:", 6)
        quick_label("Removed:", 7)
        labels = {}
        labels["ns"] = quick_label("0", 0, 1)
        labels["i"] = quick_label("0", 1, 1)
        labels["r"] = quick_label("0", 2, 1)
        labels["f"] = quick_label("0", 3, 1)
        labels["d"] = quick_label("0", 4, 1)
        labels["w"] = quick_label("0", 5, 1)
        labels["s"] = quick_label("0", 6, 1)
        labels["rm"] = quick_label("0", 1)

        frm_counter.grid(row=0, column=0)

        def tk_count() -> None:
            c = Counter()
            self._count_jobs(c)
            labels["ns"]["text"] = f"{c.notSub}"
            labels["i"]["text"] = f"{c.idle}"
            labels["r"]["text"] = f"{c.run}"
            labels["f"]["text"] = f"{c.failed}"
            labels["d"]["text"] = f"{c.done}"
            labels["w"]["text"] = f"{c.waiting}"
            labels["s"]["text"] = f"{c.skipped}"
            labels["rm"]["text"] = f"{c.removed}"

            # if no job is waiting nor running, finish the manager
            # TODO: add condition (close on finish)
            # if not (c.waiting + c.notSub + c.idle + c.run > 0):
            #    window.destroy()

            # checking dependencies and submitting ready jobs
            self._check_dependence()

            window.after(1000 * sleepTime, tk_count)

        tk_count()
        log.info("MONITOR: START")
        window.mainloop()
        log.info("MONITOR: FINISHED")

    def start(self, sleepTime: int = 60, gui: bool = False) -> None:
        """Starts the manager, iteratively checking status of jobs.

        Makes sure to save the current state of jobs
        in case of interupt or crash.

        Arguments:
            sleepTime (int, optional): time to sleep between checks.
                Defaults to 60.
            gui (bool, optional): whether to use GUI. Defaults to False.
                GUI is experimental!
        """
        try:
            if gui:
                self._start_gui(sleepTime)
            else:
                self._start_cli(sleepTime)
        except KeyboardInterrupt:
            log.error("Manager interrupted with keyboard!")
            log.error("Saving and exitting ...")
            self.save()
            self.print_failed()
            sys.exit(0)
        except Exception as e:
            log.error("Error ocurred when running manager!")
            log.error(str(e))
            traceback.print_exc(file=sys.stdout)
            self.save()
            self.print_failed()
            sys.exit(1)



from .manager import manager, Counter  # NOQA
from .job import job  # NOQA
from .status import FalconryStatus  # NOQA
from . import cli  # NOQA
from .schedd_wrapper import ScheddWrapper, kerberos_auth  # NOQA



import htcondor
import os
import logging

from typing import List, Dict, Any

from .status import FalconryStatus
from .schedd_wrapper import ScheddWrapper

log = logging.getLogger('falconry')


class job:
    """Submits and holds a single job and all relevant information

    The schedd can be imported as:

    .. code-block:: python

        from falconry import ScheddWrapper
        schedd = ScheddWrapper()

    or from the manager:

    .. code-block:: python

        from falconry import manager
        mgr = manager.Manager(mgrDir, mgrMsg)
        schedd = mgr.schedd

    Arguments:
        name (str): name of the job for easy identification
        schedd (ScheddWrapper): HTCondor schedd wrapper
    """

    def __init__(self, name: str, schedd: ScheddWrapper) -> None:

        # first, define HTCondor schedd wrapper
        self.schedd = schedd

        # name of the job for easy identification
        self.name = name

        # since we will be resubmitting, job IDs are kept as a list
        self.jobIDs: List[str] = []
        self.jobID: str = ""
        self.jobDir = ""

        # add a decoration to the job to hold dependencies
        self.dependencies: List["job"] = []

        # configuration of the jobs
        self.config: Dict[str, str] = {}

        # to setup initial state (done/submitted and so on)
        self.reset()

    def set_simple(self, exe: str, logPath: str) -> None:
        """Sets up a simple job with only executable and a path to log files

        Arguments:
            exe (str): path to the executable
            logPath (str): path to the log files
        """
        # Extend the log path
        logPathBase = os.path.abspath(logPath)
        logDir = os.path.join(logPathBase, "logs")
        self.jobDir = os.path.join(logPathBase, self.name)

        # htcondor defines job as a dict
        cfg = {
            "executable": exe,
            "log": os.path.join(logDir, "$(JobId).log"),
            "output": os.path.join(logDir, "$(JobId).out"),
            "error": os.path.join(logDir, "$(JobId).err"),
        }
        self.config = cfg

        # create the directory for the log
        if not os.path.exists(logDir):
            os.makedirs(logDir, exist_ok=True)

        if not os.path.exists(self.jobDir):
            os.makedirs(self.jobDir, exist_ok=True)

        # setup flags:
        self.reset()

    def save(self) -> Dict[str, Any]:
        """Returns a dictionary containing all relevant job information
        to be saved to a file.

        Returns:
            Dict[str, Any]: dictionary containing job information
        """
        # first rewrite dependencies using names
        depNames = [j.name for j in self.dependencies]
        jobDict = {
            "jobIDs": self.jobIDs,
            "config": self.config,
            "depNames": depNames,
            "done": "false",
        }
        # to test if job is done takes long time
        # because log file needs to be checked
        # so its best to save this status
        if self.done:
            jobDict["done"] = "true"
        return jobDict

    def load(self, jobDict: Dict[str, Any]) -> None:
        """Loads a job from a dictionary created using the save function.

        Arguments:
            jobDict (Dict[str, Any]): dictionary containing job information
        """

        # TODO: define proper "jobDict checker"
        if "jobIDs" not in jobDict.keys() and "config" not in jobDict.keys():
            log.error("Job dictionary in a wrong form")
            raise SystemError

        # the htcondor version of the configuration
        self.config = jobDict["config"]

        # setup flags:
        self.reset()
        # tmp backwards compatibility
        if "done" in jobDict and jobDict["done"] == "true":
            log.debug("Job is already done")
            self.done = True

        # set cluster IDs
        self.jobIDs = jobDict["jobIDs"]

        # if not empty, the job has been already submitted at least once
        if len(self.jobIDs) > 0:
            self.htjob = htcondor.Submit(self.config)  # type: ignore
            self.jobID = self.jobIDs[-1]

            self.expand_files()
            self.submitted = True

    def reset(self) -> None:
        """Resets job flags"""
        self.submitted = False
        self.skipped = False
        self.failed = False
        self.done = False

    def add_job_dependency(self, *args: "job") -> None:
        """Add dependencies to the job.

        Arguments:
            *args (List["job"]): list of jobs
        """
        self.dependencies.extend(list(args))

    # submit the job
    def submit(self, force: bool = False, doNotSubmit: bool = False) -> None:
        """Submits the job to HTCondor if either the job is not submitted,
        the force flag is set or the job failed.

        Arguments:
            force (bool, optional): force submission. Defaults to False.
            doNotSubmit (bool, optional): does not submit, just updates htjob. Defaults to False.
        """
        # TODO: raise error if problem

        # force: for cases when the job status was checked
        # e.g. when retrying
        # this is can save a lot of time because
        # failed job required the log file to be read.
        # Should not be used by users, only internally.

        # first check if job was not submitted before:
        if not force and self.jobIDs != []:
            status = self.get_status()
            if status in [
                FalconryStatus.ABORTED_BY_USER,
                FalconryStatus.FAILED,
                FalconryStatus.LOG_FILE_MISSING,
            ]:
                log.info("Job %s failed and will be resubmitted.", self.name)
            else:
                log.info("The job is %s, not submitting", status.name)
                return
        else:
            # the htcondor version of the configuration
            self.htjob = htcondor.Submit(self.config)  # type: ignore

        if doNotSubmit:
            return

        # Of course the submit has different capitalization here ...
        submit_result = self.schedd.submit(self.htjob)
        self.submit_done(submit_result.cluster(), "0")

    def submit_done(self, clusterID: str, procID: str) -> None:
        """Sets the job as submitted and updates cluster ID"""
        self.jobID = f"{clusterID}.{procID}"
        self.jobIDs.append(self.jobID)
        log.info(f"Submitting job {self.name} with id {self.jobID}")
        log.debug(self.config)
        self.expand_files()
        # reset job properties
        self.reset()
        self.submitted = True

    def expand_files(self) -> None:
        """Expands job files with cluster and job IDs"""

        def update_symlink(source: str, dest: str) -> None:
            if os.path.islink(dest):
                os.remove(dest)
            os.symlink(source, dest)

        self.logFile = os.path.join(self.jobDir, "last.log")
        logFile = self.config["log"].replace("$(JobId)", self.jobID)
        update_symlink(logFile, self.logFile)

        self.outFile = os.path.join(self.jobDir, "last.out")
        outFile = self.config["output"].replace("$(JobId)", self.jobID)
        update_symlink(outFile, self.outFile)

        self.errFile = os.path.join(self.jobDir, "last.err")
        errFile = self.config["error"].replace("$(JobId)", self.jobID)
        update_symlink(errFile, self.errFile)

    @property
    def clusterId(self) -> str:
        """Returns cluster ID"""
        return self.jobID.split(".")[0]

    @property
    def procId(self) -> str:
        """Returns proc ID"""
        return self.jobID.split(".")[1]

    @property
    def act_constraints(self) -> str:
        """Returns HTCondor constraints for the job"""
        return f"(ClusterId == {self.clusterId}) && (ProcId == {self.procId})"

    def release(self) -> bool:
        """Releases held job"""
        if self.jobIDs == []:
            return False
        self.schedd.act(
            htcondor.JobAction.Release, self.act_constraints  # type: ignore
        )
        log.info("Releasing job %s with id %s", self.name, self.jobID)
        return True

    def remove(self) -> bool:
        """Removes the job from HTCondor"""
        if self.jobIDs == []:
            return False
        self.schedd.act(htcondor.JobAction.Remove, self.act_constraints)  # type: ignore
        log.info("Removing job %s with id %s", self.name, self.jobID)
        return True

    def get_info(self) -> Dict[str, Any]:
        """Returns information about the job

        Returns:
            Dict[str, Any]: dictionary containing job information
        """
        # check if job has an ID
        if self.jobIDs == []:
            log.error("Trying to list info for a job which was not submitted")
            raise SystemError

        # get all job info of running job
        ads = self.schedd.query(constraint=self.act_constraints)

        # if the job finished, query will be empty and we have to use history
        # because condor is stupid, it returns and iterator (?),
        # so just returning first element
        # TODO: add more to projection
        if ads == []:
            for ad in self.schedd.history(
                constraint=self.act_constraints, projection=["JobStatus"]
            ):
                return ad

        # check if only one job was returned
        if len(ads) != 1:
            # empty job is probably job finished a long ago
            # return specific code -999 and let get_status function
            # sort the rest from the log files
            if ads == []:
                return {"JobStatus": -999}
            else:
                log.error(
                    "HTCondor returned more than one jobs for given ID, this should not happen!"
                )
                log.error(f"Job {self.name} with id {self.jobID}")
                print(ads)
                raise SystemError

        return ads[0]  # we take only single job, so return onl the first eleement

    def get_status(self) -> FalconryStatus:
        """Returns status of the job, as defined in status.py

        Returns:
            int: status of the job
        """

        # First check if the job is skipped or not even submitted
        if self.skipped:
            return FalconryStatus.SKIPPED
        elif self.done:
            return FalconryStatus.COMPLETE
        elif self.jobIDs == []:  # job was not even submitted
            return FalconryStatus.NOT_SUBMITTED
        # Using exists here is crucial, it returns false for broken symlinks
        elif not os.path.exists(self.logFile):
            return FalconryStatus.LOG_FILE_MISSING

        status_log = self._get_status_log()
        if status_log != 0:  # 0 for unknown so try from condor
            if status_log < 0:
                return FalconryStatus.FAILED
            return FalconryStatus(status_log)

        cndr_status = self._get_status_condor()
        # If job is incomplete, simply return the status:
        if cndr_status != 4 and cndr_status != -999:
            return FalconryStatus(cndr_status)

        log.error("Unknown output of job %s!", self.name)
        return FalconryStatus.UNKNOWN

    def _get_status_condor(self) -> int:
        """Returns status of the job, as defined in condor

        Returns:
            int: status of the job
        """
        return self.get_info()["JobStatus"]

    def _get_status_log(self) -> int:
        """Gets status from the log file

        Returns:
            int: status of the job
        """
        # Check log file to determine if job finished with an error
        with open(self.logFile, 'r') as fl:
            search = fl.read()

        # User abortion is special case
        if "Job was aborted by the user" in search:
            # I think this is the same as 3 but need to check
            return 12

        # Sometimes `removed` is not properly saved
        # (probably when continuing after long time?)
        # so here alternative way
        if "SYSTEM_PERIODIC_REMOVE" in search or "Job was aborted" in search:
            return 3

        # Otherwise check `"Job terminated"`. If the log does not contain it
        # its unknown state
        if "Job terminated" not in search:
            return 0

        # Evaluate `"Job terminated"`
        searchSplit = search.split("\n")
        for line in searchSplit:
            if "Normal termination (return value" in line:
                line = line.rstrip()  # remove '\n' at end of line
                status = int(line.split("value")[1].strip()[:-1])

                if status == 0:
                    self.done = True
                    return 4  # success

                log.debug(f"Job failed {status}")
                self.failed = True
                # Positive  values reserved for falconry states,
                # so return as negative
                return -status
        return 11  # no "Normal termination for Job terminated"

    def set_custom(self, dict: Dict[str, str]) -> None:
        """Sets custom configuration for the job from a dictionary

        Arguments:
            dict (Dict[str, str]): dictionary containing job configuration
        """
        for key, item in dict.items():
            self.config[key] = item

    def set_time(self, runTime: int, useRequestRuntime: bool = False) -> None:
        """Sets time limit for the job.

        For some clusters (DESY), RequestRuntime is used instead of MaxRuntime,
        to use it set useRequestRuntime to `True`.

        Arguments:
            runTime (int): time limit in seconds
            useRequestRuntime (bool, optional): use RequestRuntime option. Defaults to False.
        """
        self.config["+MaxRuntime"] = str(runTime)
        # RequestRuntime seems to be DESY specific option and does not work
        # e.g. in Prague (jobs get held), so this option is false by default
        if useRequestRuntime:
            self.config["+RequestRuntime"] = str(runTime)

    def set_arguments(self, args: str) -> None:
        """Sets arguments for the job

        Arguments:
            args (str): arguments for the job
        """
        self.config["arguments"] = args



from enum import Enum


class FalconryStatus(Enum):
    UNKNOWN = 0
    IDLE = 1
    RUNNING = 2
    REMOVED = 3
    COMPLETE = 4
    HELD = 5
    TRANSPORTING = 6
    SUSPENDED = 7
    SKIPPED = 8
    NOT_SUBMITTED = 9
    LOG_FILE_MISSING = 10
    ABNORMAL_TERMINATION = 11
    ABORTED_BY_USER = 12
    FAILED = 13
